
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "sif.h"

extern void	Error(),RangeList(),ReadSliceImage();

#define PV	(Cell)(~0)
#define LEN	1024
#define BPS	(sizeof(Cell)<<3)

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,pv[(size_t)PV+1],**last,form[LEN],path[LEN],
		*moa="missing object area.";
	int	Nx,Ny,Nz,Ne,Nd,Nl,Nxl,Nyl,Nzl,Lx,Ly,Lz,
		Z,x,y,z,x1,y1,z1,x2,y2,z2,l,x3,y3,z3,x4,y4,z4,X,Y,len;
	Cell	***image,**cell,**block,mask,**Block,Mask;

	if (argc!=6 && argc!=7)
    Error("usage : si_eda orgDir nameFile rangeList erosion dilation {newDir}");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],(size_t)PV,pv);

	if ((Ne=atoi(argv[4]))<0 ||
	    (Nd=atoi(argv[5]))<0) Error("bad layers.");

	Nl=(Nd>Ne)?Nd-Ne:0; Nxl=Nx+Nl; Lx=Nl+Nxl;
			    Nyl=Ny+Nl; Ly=Nl+Nyl;
			    Nzl=Nz+Nl; Lz=Nl+Nzl;

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

	Z=(Lz+BPS-1)/BPS;

	if ((image=ALLOC(Cell **,Z))==NULL ||
	    (*image=ALLOC(Cell *,(size_t)Z*(size_t)Ly))==NULL ||
	    (**image=ALLOC(Cell,(size_t)Z*(size_t)Ly*(size_t)Lx))==NULL)
	    Error("no allocatable memory for image.");

	cell=image[0]; for (x=0; x<Lx; x++) cell[0][x]=0;
	for (y=1; y<Ly; y++) {
	    cell[y]=cell[y-1]+Lx; for (x=0; x<Lx; x++) cell[y][x]=0;
	}
	for (z=1; z<Z; z++) {
	    cell=image[z]=image[z-1]+Ly;
	    for (y=0; y<Ly; y++) {
		cell[y]=cell[y-1]+Lx; for (x=0; x<Lx; x++) cell[y][x]=0;
	    }
	}
	if ((cell=ALLOC(Cell *,Ly))==NULL ||
	    (*cell=ALLOC(Cell,(size_t)Ly*(size_t)Lx))==NULL)
	    Error("no allocatable memory for cell.");

	for (y=1; y<Ly; y++) cell[y]=cell[y-1]+Lx;

	x1=Nxl; y1=Nyl; z1=Nzl; x2=y2=z2=Nl-1;
	for (z=Nl; z<Nzl; z++) {
	    ReadSliceImage(paths[z-Nl],Nx,Ny,cell);
	    block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
	    for (y=Nl; y<Nyl; y++) for (x=Nl; x<Nxl; x++)
		if (pv[cell[y-Nl][x-Nl]]) {
		    block[y][x]|=mask; if (x<x1) x1=x; if (x>x2) x2=x;
				       if (y<y1) y1=y; if (y>y2) y2=y;
				       if (z<z1) z1=z; if (z>z2) z2=z;
		}
	}
	if (x1>x2 || y1>y2 || z1>z2) Error("no object area.");

	(void)printf("%d\t%d\t%d\t%d\t%d\t%d\n",
		     x1-Nl,y1-Nl,z1-Nl,x2-Nl,y2-Nl,z2-Nl);

	l=Ne*2; if (x2-x1<=l || y2-y1<=l || z2-z1<=l) Error(moa);

	if (argc!=7) {
	    free(*cell); free(cell);
	}
	if ((last=ALLOC(char *,1+Ly))==NULL ||
	    (*last=ALLOC(char,(size_t)(1+Ly)*(size_t)(1+Lx)))==NULL)
	    Error("no allocatable memory for last.");

	++last; ++last[-1];
	for (y=0; y<Ly; y++) {
	    last[y]=last[y-1]+Lx; ++last[y];
	}
	for (l=0; l<Ne; l++) {

#include "e3.h"

	    if (x1>x2 || y1>y2 || z1>z2) Error(moa);
	}
	(void)printf("%d\t%d\t%d\t%d\t%d\t%d\n",
		     x1-Nl,y1-Nl,z1-Nl,x2-Nl,y2-Nl,z2-Nl);

	for (l=0; l<Nd; l++) {

#include "d3.h"

	}
	(void)printf("%d\t%d\t%d\t%d\t%d\t%d\n",
		     x1-Nl,y1-Nl,z1-Nl,x2-Nl,y2-Nl,z2-Nl);

	if (argc==7) {
	    X=x2-x1+1; Y=y2-y1+1; Z=z2-z1+1;
	    len=1; for (z=10; z<Z; z*=10) ++len;
	    (void)sprintf(form,"%s/%%0%dd.tif",argv[6],len);

	    for (z=z1; z<=z2; z++) {
		block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
		for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
		    cell[y-y1][x-x1]=(block[y][x]&mask)?1:0;

		(void)sprintf(path,form,z-z1);
		StoreImageFile(path,X,Y,1,cell,NULL);
	    }
	}
	return 0;
}
