
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "sif.h"

#ifndef	LABEL_T
#define LABEL_T	unsigned int
#endif

extern void	Error(),RangeList(),ReadSliceImage();

static LABEL_T	Trace(link,old)
LABEL_T		*link,old;
{
	LABEL_T	new=link[old];

	return (new==0)?old:(link[old]=Trace(link,new));
}

#define PV	(Cell)(~0)
#define LEN	1024
#define BPS	(sizeof(Cell)<<3)

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,pv[(size_t)PV+1],path[LEN],*moa="missing object area.";
	int	Nx,Ny,Nz,Nt,Z,x,y,z,x1,y1,z1,x2,y2,z2,t,x3,y3,z3,x4,y4,z4;
	Cell	***image,**cell,**block,mask,**Block,Mask;
	LABEL_T	**last,Nl,*link,l,lxy;
	size_t	len;

	if (argc!=5 && argc!=6)
	Error("usage : si_osp3 orgDir nameFile rangeList thickness {newDir}");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],(size_t)PV,pv);

	if ((Nt=atoi(argv[4]))<0) Error("bad thickness.");

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

	Z=(Nz+BPS-1)/BPS;

	if ((image=ALLOC(Cell **,Z))==NULL ||
	    (*image=ALLOC(Cell *,(size_t)Z*(size_t)Ny))==NULL ||
	    (**image=ALLOC(Cell,(size_t)Z*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for image.");

	cell=image[0]; for (x=0; x<Nx; x++) cell[0][x]=0;
	for (y=1; y<Ny; y++) {
	    cell[y]=cell[y-1]+Nx; for (x=0; x<Nx; x++) cell[y][x]=0;
	}
	for (z=1; z<Z; z++) {
	    cell=image[z]=image[z-1]+Ny;
	    for (y=0; y<Ny; y++) {
		cell[y]=cell[y-1]+Nx; for (x=0; x<Nx; x++) cell[y][x]=0;
	    }
	}
	if ((cell=ALLOC(Cell *,Ny))==NULL ||
	    (*cell=ALLOC(Cell,(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for cell.");

	for (y=1; y<Ny; y++) cell[y]=cell[y-1]+Nx;

	x1=Nx; y1=Ny; z1=Nz; x2=y2=z2=(-1);
	for (z=0; z<Nz; z++) {
	    ReadSliceImage(paths[z],Nx,Ny,cell);
	    block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++)
		if (pv[cell[y][x]]) {
		    block[y][x]|=mask; if (x<x1) x1=x; if (x>x2) x2=x;
				       if (y<y1) y1=y; if (y>y2) y2=y;
				       if (z<z1) z1=z; if (z>z2) z2=z;
		}
	}
	if (x1>x2 || y1>y2 || z1>z2) Error("no object area.");

	(void)printf("%d\t%d\t%d\t%d\t%d\t%d\n",x1,y1,z1,x2,y2,z2);

	t=Nt*2; if (x2-x1<=t || y2-y1<=t || z2-z1<=t) Error(moa);

	if (argc!=6) {
	    free(*cell); free(cell);
	}
	if ((last=ALLOC(LABEL_T *,1+Ny))==NULL ||
	    (*last=ALLOC(LABEL_T,(size_t)Nx+(size_t)Ny*(size_t)(1+Nx)))==NULL)
	    Error("no allocatable memory for last.");

	++last;
	for (x=0; x<Nx; x++) last[-1][x]=1;
	for (y=0; y<Ny; y++) {
	    last[y]=last[y-1]+1+Nx; for (x=(-1); x<Nx; x++) last[y][x]=1;
	}
	for (Nl=2, z=0; z<Nz; z++) {
	    block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++)
		if (block[y][x]&mask)
		    last[y][x]=0;
		else {
		    if (!last[y][x] && !last[y-1][x] && !last[y][x-1])
			if (++Nl>(LABEL_T)(~0)/sizeof(LABEL_T))
			    Error("too many links.");

		    last[y][x]=1; 
		}
	}
	if ((link=ALLOC(LABEL_T,Nl-1))==NULL)
	    Error("no allocatable memory for link.");

	--link; link[1]=0;

	for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++) last[y][x]=1;

	for (Nl=2, z=z1; z<=z2; z++) {
	    block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
	    for (y=y1; y<=y2; y++) {
		for (x=x1; x<=x2; x++)
		    if (block[y][x]&mask)
			last[y][x]=0;
		    else if ((l=last[y][x])!=0) {
			l=Trace(link,l);
			if ((lxy=last[y-1][x])!=0 && (lxy=Trace(link,lxy))!=l)
			    if (lxy==1) l=link[l]=1; else link[lxy]=l;

			if ((lxy=last[y][x-1])!=0 && (lxy=Trace(link,lxy))!=l)
			    if (lxy==1) l=link[l]=1; else link[lxy]=l;

			last[y][x]=l;
		    }
		    else if ((l=last[y-1][x])!=0) {
			l=Trace(link,l);
			if ((lxy=last[y][x-1])!=0 && (lxy=Trace(link,lxy))!=l)
			    if (lxy==1) l=link[l]=1; else link[lxy]=l;

			last[y][x]=l;
		    }
		    else if ((l=last[y][x-1])!=0)
			last[y][x]=Trace(link,l);
		    else
			link[last[y][x]=Nl++]=0;

		if (z>z1 && y>y1 && x2>x1)
		    if ((l=last[y][x2])!=0 && (l=Trace(link,l))!=1) link[l]=1;
	    }
	    if (z>z1 && y2>y1)
		for (x=x1+1; x<x2; x++)
		    if ((l=last[y2][x])!=0 && (l=Trace(link,l))!=1) link[l]=1;
	}
	if (z2>z1)
	    for (y=y1+1; y<y2; y++) for (x=x1+1; x<x2; x++)
		if ((l=last[y][x])!=0 && (l=Trace(link,l))!=1) link[l]=1;

	for (l=2; l<Nl; l++) if (link[l]!=0) (void)Trace(link,l);
	for (l=2; l<Nl; l++) if (link[l]==0) link[l]=l;

	for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++) last[y][x]=1;

	for (Nl=2, z=z1; z<=z2; z++) {
	    block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
	    for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
		if (block[y][x]&mask)
		    last[y][x]=0;
		else {
		    if ((l=last[y  ][x  ])==0 &&
			(l=last[y-1][x  ])==0 &&
			(l=last[y  ][x-1])==0) l=link[Nl++];

		    if ((last[y][x]=l)!=1) block[y][x]^=mask;
		}
	}
	free(link+1);

	for (t=0; t<Nt; t++) {

#include "e3.h"

	    if (x1>x2 || y1>y2 || z1>z2) Error(moa);
	}
	(void)printf("%d\t%d\t%d\t%d\t%d\t%d\n",x1,y1,z1,x2,y2,z2);

	if (argc==6) {
	    len=strlen(argv[1])+1;
	    for (z=0; z<Nz; z++) {
		block=image[z/BPS]; mask=(Cell)(1<<z%BPS);
		for (y=0; y<Ny; y++) for (x=0; x<Nx; x++)
		    cell[y][x]=(block[y][x]&mask)?1:0;

		(void)sprintf(path,"%s/%s",argv[5],paths[z]+len);
		StoreImageFile(path,Nx,Ny,1,cell,NULL);
	    }
	}
	return 0;
}
